//! This module contains the Cap'n Proto encoding and decoding functions
//! generated from the `.capn` schema.
//!
//! These files **SHOULD NOT** be edited manually.
//! Instead, use `just update-model-capnp` to regenerate them.

#[allow(warnings)]
#[allow(unknown_lints)]
#[allow(clippy::all)]
#[rustfmt::skip]
pub mod hugr_v0_capnp;
